import json
import math
import random

import torch.nn.functional as F
from sklearn.decomposition import PCA
from scipy import spatial

import torch.utils
import torch
import numpy as np
import matplotlib.pyplot as plt
import torchvision
from sklearn.preprocessing import MinMaxScaler

BATCH_SIZE = 16
LEARNING_RATE = 1e-4

HIDDEN_SIZE = 128
LSTM_LAYERS = 1

INPUT_FEATURES = 2

DEVICE = 'cpu'
MAX_LEN = 200
if torch.cuda.is_available():
    MAX_LEN = None
    DEVICE = 'cuda'

CLASSES_INCLUDED = [0, 1] #torch.arange(10).tolist() # Inlude more class
OUTPUT_CLASSES = len(CLASSES_INCLUDED)

class DatasetMNIST(torch.utils.data.Dataset):
    def __init__(self, is_train):
        self.data_raw = torchvision.datasets.MNIST( # EMNIST
            root='./data',
            train=is_train,
            download=True
        )
        self.scaler_x = MinMaxScaler(feature_range=(-1, 1))
        self.data = []
        for x, y_idx in self.data_raw:
            if y_idx in CLASSES_INCLUDED: # select only 3 clases form datasets
                if MAX_LEN and len(self.data) > MAX_LEN:
                    break
                self.data.append((x, y_idx))
                self.scaler_x.partial_fit(np.array(x))

    def __len__(self):
        if MAX_LEN:
            return MAX_LEN
        return len(self.data)

    def __getitem__(self, idx):
        pil_x, y_idx = self.data[idx]
        y_idx = CLASSES_INCLUDED.index(y_idx)
        np_x = np.array(pil_x) # (28, 28)
        np_x = self.scaler_x.transform(np_x)
        np_x = np_x.reshape(int(28*28/INPUT_FEATURES), INPUT_FEATURES)
        x = torch.FloatTensor(np_x)
        y_idx = torch.LongTensor(np.array([y_idx]))
        return x, y_idx

data_loader_train = torch.utils.data.DataLoader(
    dataset=DatasetMNIST(is_train=True),
    batch_size=BATCH_SIZE,
    shuffle=True,
    drop_last=True
)
data_loader_test = torch.utils.data.DataLoader(
    dataset=DatasetMNIST(is_train=False),
    batch_size=BATCH_SIZE,
    shuffle=False,
    drop_last=True
)

class LSTM(torch.nn.Module):
    def __init__(self, input_size, hidden_size):
        super().__init__()

        self.hidden_size = hidden_size
        stdv = 1.0 / math.sqrt(self.hidden_size)

        self.W = torch.nn.Parameter(
            torch.FloatTensor(4 * input_size, 4 * hidden_size).uniform_(-stdv, stdv) # Test -1..1, default 0..1
        )

        self.U = torch.nn.Parameter(
            torch.FloatTensor(4 * input_size, 4 * hidden_size).uniform_(-stdv, stdv) # Test -1..1, default 0..1
        )

        self.b = torch.nn.Parameter(
            torch.FloatTensor(4 * hidden_size).zero_()
        ) #TODO init forget gate part as 1.0 (self.hidden_size:self.hidden_size*2)

        self.w_peep = torch.nn.Parameter(
            torch.FloatTensor(3 * hidden_size).uniform_(-stdv, stdv)
        )

    def forward(self, x, h_c=None):
        if h_c is None:
            h = torch.zeros((x.size(0), self.hidden_size)).to(DEVICE) # TODO learnable h_0
            c = torch.zeros((x.size(0), self.hidden_size)).to(DEVICE)
        else:
            h, c = h_c
        h_out = []
        # cuda dnn
        # x => (B, Seq, F)
        x_seq = x.permute(1, 0, 2) # (Seq, B, F)
        for x_t in x_seq:

            gates = torch.matmul(
                x_t.repeat(1, 3),
                self.W[:self.hidden_size*3, :self.hidden_size*3] # (in, out)
            ) + torch.matmul(
                h.repeat(1, 3),
                self.U[:self.hidden_size*3, :self.hidden_size*3]
            ) + self.b[:self.hidden_size*3] + self.w_peep * c.repeat(1, 3) # ? should this be c_t or/and c_{t-1}

            i_t = torch.sigmoid(gates[:, 0:self.hidden_size])
            f_t = torch.sigmoid(gates[:, self.hidden_size:self.hidden_size*2])
            o_t = torch.sigmoid(gates[:, self.hidden_size*2:self.hidden_size*3])

            gate_c = torch.matmul(
                x_t,
                self.W[self.hidden_size*3:self.hidden_size*4, self.hidden_size*3:self.hidden_size*4]
            ) + torch.matmul(
                h,
                self.U[self.hidden_size*3:self.hidden_size*4, self.hidden_size*3:self.hidden_size*4]
            ) + self.b[self.hidden_size*3:self.hidden_size*4]

            c = f_t * c + i_t * torch.tanh(gate_c)
            h = torch.tanh(c) * o_t
            h_out.append(h)
        t_h_out = torch.stack(h_out)
        t_h_out = t_h_out.permute(1, 0, 2) #  (B, Seq, F)
        return t_h_out


class ModelLSTM(torch.nn.Module):
    def __init__(self):
        super().__init__()

        self.ff = torch.nn.Linear(
            in_features=INPUT_FEATURES,
            out_features=HIDDEN_SIZE
        )

        layers = []
        for _ in range(LSTM_LAYERS):
            layers.append(LSTM(
                input_size=HIDDEN_SIZE,
                hidden_size=HIDDEN_SIZE,
            ))
        self.lstm = torch.nn.Sequential(*layers)

        self.fc = torch.nn.Linear(
            in_features=HIDDEN_SIZE,
            out_features=OUTPUT_CLASSES
        )

    def forward(self, x):
        # B, Seq, F => B, 28, 28
        z_0 = self.ff.forward(x)
        z_1 = self.lstm.forward(z_0) # B, Seq, Hidden_size
        z_2 = torch.mean(z_1, dim=1).squeeze() # B, Hidden_size
        logits = self.fc.forward(z_2)
        y_prim = torch.softmax(logits, dim=1)
        return y_prim

model = ModelLSTM()
model = model.to(DEVICE)

optimizer = torch.optim.Adam(model.parameters(), lr=LEARNING_RATE)

metrics = {}
for stage in ['train', 'test']:
    for metric in [
        'loss',
        'acc'
    ]:
        metrics[f'{stage}_{metric}'] = []

for epoch in range(1, 100):

    for data_loader in [data_loader_train, data_loader_test]:
        metrics_epoch = {key: [] for key in metrics.keys()}

        stage = 'train'
        if data_loader == data_loader_test:
            stage = 'test'
            model = model.eval()
        else:
            model = model.train()

        for x, y_idx in data_loader:
            x = x.to(DEVICE)
            y_idx = y_idx.to(DEVICE).squeeze()

            y_prim = model.forward(x)

            idxes = torch.arange(x.size(0)).to(DEVICE)
            loss = -torch.mean(torch.log(y_prim[idxes, y_idx] + 1e-8))
            y_idx_prim = torch.argmax(y_prim, dim=1)

            acc = torch.mean((y_idx == y_idx_prim) * 1.0)
            metrics_epoch[f'{stage}_loss'].append(loss.cpu().item())
            metrics_epoch[f'{stage}_acc'].append(acc.cpu().item())

            if data_loader == data_loader_train:
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()

        metrics_strs = []
        for key in metrics_epoch.keys():
            if stage in key:
                value = np.mean(metrics_epoch[key])
                metrics[key].append(value)
                metrics_strs.append(f'{key}: {round(value, 2)}')

        print(f'epoch: {epoch} stage: {stage} {" ".join(metrics_strs)}')

    plts = []
    c = 0
    for key, value in metrics.items():
        plts += plt.plot(value, f'C{c}', label=key)
        ax = plt.twinx()
        c += 1
    plt.legend(plts, [it.get_label() for it in plts])
    plt.show()